"""Tests for lakshmi module."""
import unittest
from unittest.mock import patch

import lakshmi.cache
from lakshmi import Account, AssetClass, Portfolio
from lakshmi.assets import ManualAsset, TaxLot, TickerAsset
from lakshmi.table import Table


class LakshmiTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        lakshmi.cache.set_cache_dir(None)  # Disable caching.

    def test_empty_portfolio(self):
        portfolio = Portfolio(AssetClass('E'))
        self.assertAlmostEqual(0, portfolio.total_value())
        self.assertListEqual([], portfolio.assets().list())
        self.assertListEqual([], portfolio.asset_location().list())
        self.assertListEqual([], portfolio.asset_allocation_tree().list())
        self.assertListEqual([], portfolio.asset_allocation([]).list())
        self.assertListEqual([], portfolio.asset_allocation_compact().list())
        self.assertListEqual([], portfolio.list_lots().list())

    def test_one_asset_class(self):
        AssetClass('Equity').validate()

    def test_many_asset_class_duplicate(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('International')))
            .add_subclass(0.2, AssetClass('US')))
        with self.assertRaisesRegex(AssertionError, 'Found duplicate'):
            asset_class.validate()

    def test_many_asset_class_bad_ratio_sum(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.5, AssetClass('International')))
            .add_subclass(0.2, AssetClass('Bonds')))

        with self.assertRaisesRegex(AssertionError, 'Sum of sub-classes'):
            asset_class.validate()

    def test_many_asset_class_bad_ratio_neg(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(-0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('International')))
            .add_subclass(0.2, AssetClass('Bonds')))

        with self.assertRaisesRegex(AssertionError, 'Bad ratio'):
            asset_class.validate()

    def test_many_asset_class_bad_ratio_high(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(1.5,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('International')))
            .add_subclass(0.2, AssetClass('Bonds')))

        with self.assertRaisesRegex(AssertionError, 'Bad ratio'):
            asset_class.validate()

    def test_many_asset_class(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('International')))
            .add_subclass(0.2, AssetClass('Bonds'))).validate()

        self.assertEqual(
            {'US', 'International', 'Bonds'},
            asset_class.leaves())

        ret_class, ratio = asset_class.find_asset_class('All')
        self.assertEqual('All', ret_class.name)
        self.assertAlmostEqual(1.0, ratio)
        ret_class, ratio = asset_class.find_asset_class('Equity')
        self.assertEqual('Equity', ret_class.name)
        self.assertAlmostEqual(0.8, ratio)
        ret_class, ratio = asset_class.find_asset_class('US')
        self.assertEqual('US', ret_class.name)
        self.assertAlmostEqual(0.48, ratio)
        ret_class, ratio = asset_class.find_asset_class('International')
        self.assertEqual('International', ret_class.name)
        self.assertAlmostEqual(0.32, ratio)
        ret_class, ratio = asset_class.find_asset_class('Bonds')
        self.assertEqual('Bonds', ret_class.name)
        self.assertAlmostEqual(0.2, ratio)

    def test_asset_class_dict(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('International')))
            .add_subclass(0.2, AssetClass('Bonds'))).validate()

        asset_class = AssetClass.from_dict(asset_class.to_dict())
        self.assertEqual(
            {'US', 'International', 'Bonds'},
            asset_class.leaves())
        ret_class, ratio = asset_class.find_asset_class('US')
        self.assertEqual('US', ret_class.name)
        self.assertAlmostEqual(0.48, ratio)

    def test_asset_class_copy(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('International')))
            .add_subclass(0.2, AssetClass('Bonds'))).validate()

        asset_class2 = asset_class.copy().validate()
        asset_class2.name = 'Changed'
        self.assertEqual('All', asset_class.name)

    def test_value_mapped(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(0.6, AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.4, AssetClass('Bonds')))

        with self.assertRaisesRegex(AssertionError, 'Need to validate'):
            asset_class.value_mapped({})

        asset_class.validate()
        money_allocation = {
            'US': 10.0,
            'Intl': 20.0,
            'Bonds': 40.0,
            'Unused': 50.0}

        self.assertAlmostEqual(
            70.0, asset_class.value_mapped(money_allocation))
        self.assertAlmostEqual(
            30.0, asset_class.children()[0][0].value_mapped(money_allocation))
        self.assertAlmostEqual(
            40.0, asset_class.children()[1][0].value_mapped(money_allocation))

    def test_bad_asset(self):
        portfolio = Portfolio(AssetClass('Equity'))
        account = Account('Roth IRA', 'Post-tax').add_asset(
            ManualAsset('Test Asset', 100.0, {'Bad Equity': 1.0}))
        with self.assertRaisesRegex(
                AssertionError,
                'Unknown or non-leaf asset class: Bad Equity'):
            portfolio.add_account(account)

    def test_get_set_assets_from_account(self):
        account = (Account('Roth IRA', 'Post-tax')
                   .add_asset(ManualAsset('Test 1', 100.0, {'All': 1.0}))
                   .add_asset(ManualAsset('Test 2', 200.0, {'All': 1.0})))
        asset = account.get_asset('Test 1')
        self.assertEqual('Test 1', asset.name())
        self.assertAlmostEqual(100.0, asset.value())

        account.add_asset(ManualAsset('Test 1', 300.0, {'All': 1.0}),
                          replace=True)
        account.set_assets(account.assets())
        self.assertEqual(2, len(account.assets()))
        account.remove_asset('Test 1')
        with self.assertRaises(KeyError):
            account.get_asset('Test 1')

    def test_account_dict(self):
        account = Account('Roth IRA', 'Post-tax').add_asset(
            ManualAsset('Test Asset', 100.0, {'All': 1.0}))
        account.add_cash(200)
        account = Account.from_dict(account.to_dict())
        self.assertEqual(1, len(account._assets))
        self.assertEqual('Test Asset', account.get_asset('Test Asset').name())
        self.assertAlmostEqual(200.0, account.available_cash())

    def test_account_string(self):
        account = Account('Roth IRA', 'Post-tax').add_asset(
            ManualAsset('Test', 100.0, {'All': 1.0}))
        expected = (Table(2)
                    .add_row(['Name:', 'Roth IRA'])
                    .add_row(['Type:', 'Post-tax'])
                    .add_row(['Total:', '$100.00']))
        self.assertEqual(expected.string(tablefmt='plain'), account.string())

        account.add_cash(-10)
        expected.add_row(['Available Cash:', '-$10.00'])
        self.assertEqual(expected.string(tablefmt='plain'), account.string())

    def test_duplicate_account(self):
        portfolio = Portfolio(AssetClass('All'))
        account = Account('Roth IRA', 'Post-tax').add_asset(
            ManualAsset('Test Asset', 100.0, {'All': 1.0}))
        portfolio.add_account(account)
        with self.assertRaisesRegex(AssertionError, 'Attempting to add'):
            portfolio.add_account(account)
        portfolio.add_account(account, replace=True)

    def test_remove_account(self):
        portfolio = Portfolio(AssetClass('All'))
        account = Account('Roth IRA', 'Post-tax').add_asset(
            ManualAsset('Test Asset', 100.0, {'All': 1.0}))
        portfolio.add_account(account)
        portfolio.remove_account('Roth IRA')
        self.assertEqual(0, len(portfolio.accounts()))

    def test_one_asset(self):
        portfolio = Portfolio(AssetClass('Equity')).add_account(
            Account('401(k)', 'Pre-tax').add_asset(
                ManualAsset('Test Asset', 100.0, {'Equity': 1.0})))

        self.assertEqual(1, len(portfolio.accounts()))
        self.assertListEqual([['401(k)', 'Test Asset', '$100.00']],
                             portfolio.assets().str_list())
        self.assertAlmostEqual(100.0, portfolio.total_value())
        self.assertListEqual([['Equity', 'Pre-tax', '100.0%', '$100.00']],
                             portfolio.asset_location().str_list())
        self.assertListEqual([], portfolio.asset_allocation_tree().list())
        self.assertListEqual(
            [['Equity', '100.0%', '100.0%', '$100.00', '+$0.00']],
            portfolio.asset_allocation(['Equity']).str_list())
        self.assertListEqual([], portfolio.asset_allocation_tree().list())
        self.assertListEqual([], portfolio.asset_allocation_compact().list())

    def test_portfolio_dict(self):
        portfolio = Portfolio(AssetClass('Equity')).add_account(
            Account('401(k)', 'Pre-tax').add_asset(
                ManualAsset('Test Asset', 100.0, {'Equity': 1.0})))
        portfolio = Portfolio.from_dict(portfolio.to_dict())
        self.assertEqual(1, len(portfolio.accounts()))

    def test_asset_what_if(self):
        asset = ManualAsset('Test Asset', 100.0, {'Equity': 1.0})
        self.assertAlmostEqual(100, asset.adjusted_value())
        asset.what_if(-10.0)
        self.assertAlmostEqual(100, asset.value())
        self.assertAlmostEqual(90, asset.adjusted_value())
        asset.what_if(10)
        self.assertAlmostEqual(100, asset.adjusted_value())

    def test_one_asset_two_class(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.5, AssetClass('Equity'))
            .add_subclass(0.5, AssetClass('Fixed Income'))).add_account(
            Account('Vanguard', 'Taxable').add_asset(
                ManualAsset('Test Asset', 100.0,
                            {'Equity': 0.6, 'Fixed Income': 0.4})))

        self.assertListEqual([['Vanguard', 'Test Asset', '$100.00']],
                             portfolio.assets().str_list())
        self.assertAlmostEqual(100.0, portfolio.total_value())
        self.assertListEqual(
            [['Equity', 'Taxable', '100.0%', '$60.00'],
             ['Fixed Income', 'Taxable', '100.0%', '$40.00']],
            portfolio.asset_location().str_list())

        self.assertListEqual(
            [['All:'],
             ['Equity', '60.0%', '50.0%', '$60.00'],
                ['Fixed Income', '40.0%', '50.0%', '$40.00']],
            portfolio.asset_allocation_tree().str_list())
        self.assertListEqual(
            [['Equity', '60.0%', '50.0%', '$60.00', '-$10.00'],
             ['Fixed Income', '40.0%', '50.0%', '$40.00', '+$10.00']],
            portfolio.asset_allocation(['Equity', 'Fixed Income']).str_list())
        self.assertListEqual([
            ['Equity', '60%', '50%', '60.0%', '50.0%', '$60.00', '-$10.00'],
            ['Fixed Income', '40%', '50%', '40.0%', '50.0%', '$40.00',
             '+$10.00']],
            portfolio.asset_allocation_compact().str_list())

    def test_list_accounts_no_money(self):
        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable'))

        self.assertListEqual(
            [['Schwab', 'Taxable', '$0.00']],
            portfolio.list_accounts().str_list())
        self.assertListEqual(
            [['Taxable', '$0.00']],
            portfolio.list_accounts(group_by_type=True).str_list())

    def test_list_accounts(self):
        portfolio = (Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable').add_asset(
                ManualAsset('Fund', 100.0, {'All': 1.0})))
            .add_account(Account('Vanguard', 'Taxable').add_asset(
                ManualAsset('Fund', 50.0, {'All': 1.0})))
            .add_account(Account('Fidelity', '401K').add_asset(
                ManualAsset('F', 50.0, {'All': 1.0}))))

        self.assertListEqual(
            [['Schwab', 'Taxable', '$100.00', '50.0%'],
             ['Vanguard', 'Taxable', '$50.00', '25.0%'],
             ['Fidelity', '401K', '$50.00', '25.0%']],
            portfolio.list_accounts().str_list())
        self.assertListEqual(
            [['Taxable', '$150.00', '75.0%'],
             ['401K', '$50.00', '25.0%']],
            portfolio.list_accounts(group_by_type=True).str_list())

    @patch('lakshmi.assets.TickerAsset.name')
    @patch('lakshmi.assets.TickerAsset.price')
    def test_list_assets(self, mock_price, mock_name):
        mock_price.return_value = 1.0
        mock_name.return_value = 'Vanguard Cash Reserves Federal'

        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable')
            .add_asset(TickerAsset('VMMXX', 420.0, {'All': 1.0}))
            .add_asset(ManualAsset('Cash', 840.0, {'All': 1.0})))

        self.assertListEqual(
            [['Schwab', 'Vanguard Cash Reserves Federal', '$420.00'],
             ['Schwab', 'Cash', '$840.00']],
            portfolio.assets().str_list())
        self.assertListEqual(
            [['Schwab', 'VMMXX', 'Vanguard Cash Reserves Federal', '$420.00'],
             ['Schwab', 'Cash', 'Cash', '$840.00']],
            portfolio.assets(short_name=True).str_list())
        self.assertListEqual(
            [['Schwab', 'VMMXX', '420.0', 'Vanguard Cash Reserves Federal',
              '$420.00'],
             ['Schwab', 'Cash', '', 'Cash', '$840.00']],
            portfolio.assets(short_name=True, quantity=True).str_list())
        self.assertListEqual(
            [['Schwab', '420.0', 'Vanguard Cash Reserves Federal', '$420.00'],
             ['Schwab', '', 'Cash', '$840.00']],
            portfolio.assets(quantity=True).str_list())
        self.assertListEqual(
            [['Schwab', 'VMMXX', '$420.00'],
             ['Schwab', 'Cash', '$840.00']],
            portfolio.assets(short_name=True, long_name=False).str_list())

    def test_asset_location(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.2, AssetClass('Bonds')).validate())
        (portfolio
         .add_account(Account('Account1', 'Taxable')
                      .add_asset(ManualAsset('US A', 60.0, {'US': 1.0}))
                      .add_asset(ManualAsset('Intl A', 30.0, {'Intl': 1.0}))
                      .add_asset(ManualAsset('Bond A', 10.0, {'Bonds': 1.0})))
         .add_account(Account('Account2', 'Pre-tax')
                      .add_asset(ManualAsset('Bond A', 40.0, {'Bonds': 1.0}))))
        self.assertEqual(
            [['US', 'Taxable', '100.0%', '$60.00'],
             ['Intl', 'Taxable', '100.0%', '$30.00'],
             ['Bonds', 'Pre-tax', '80.0%', '$40.00'],
             ['', 'Taxable', '20.0%', '$10.00']],
            portfolio.asset_location().str_list())

    def test_flat_asset_allocation(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.2, AssetClass('Bonds')).validate()).add_account(
            Account('Account', 'Taxable')
            .add_asset(ManualAsset('US Asset', 60.0, {'US': 1.0}))
            .add_asset(ManualAsset('Intl Asset', 30.0, {'Intl': 1.0}))
            .add_asset(ManualAsset('Bond Asset', 10.0, {'Bonds': 1.0})))

        with self.assertRaisesRegex(AssertionError,
                                    'AssetAllocation called with'):
            portfolio.asset_allocation(['Equity', 'Intl'])

        self.assertListEqual(
            [['US', '60.0%', '48.0%', '$60.00', '-$12.00'],
             ['Intl', '30.0%', '32.0%', '$30.00', '+$2.00'],
             ['Bonds', '10.0%', '20.0%', '$10.00', '+$10.00']],
            portfolio.asset_allocation(['US', 'Intl', 'Bonds']).str_list())
        self.assertListEqual(
            [['Equity', '90.0%', '80.0%', '$90.00', '-$10.00'],
             ['Bonds', '10.0%', '20.0%', '$10.00', '+$10.00']],
            portfolio.asset_allocation(['Equity', 'Bonds']).str_list())

    def test_asset_allocation_compact(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.8,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.2, AssetClass('Bonds')).validate()).add_account(
            Account('Account', 'Taxable')
            .add_asset(ManualAsset('US Asset', 60.0, {'US': 1.0}))
            .add_asset(ManualAsset('Intl Asset', 30.0, {'Intl': 1.0}))
            .add_asset(ManualAsset('Bond Asset', 10.0, {'Bonds': 1.0})))

        self.assertListEqual(
            [['Equity', '90%', '80%', 'US', '67%', '60%', '60.0%', '48.0%',
              '$60.00', '-$12.00'],
             ['', '', '', 'Intl', '33%', '40%', '30.0%', '32.0%',
              '$30.00', '+$2.00'],
             ['Bonds', '10%', '20%', '', '', '', '10.0%', '20.0%',
              '$10.00', '+$10.00']],
            portfolio.asset_allocation_compact().str_list())
        self.assertListEqual(
            [['All:'],
             ['Equity', '90.0%', '80.0%', '$90.00'],
             ['Bonds', '10.0%', '20.0%', '$10.00'],
             [' '],
             ['Equity:'],
             ['US', '66.7%', '60.0%', '$60.00'],
             ['Intl', '33.3%', '40.0%', '$30.00']],
            portfolio.asset_allocation_tree().str_list())

    def test_multiple_accounts_and_assets(self):
        portfolio = Portfolio(AssetClass('All'))
        asset_class_map = {'All': 1.0}
        (portfolio
         .add_account(
             Account('Account 1', 'Taxable')
             .add_asset(ManualAsset('Asset 1', 100.0, asset_class_map))
             .add_asset(ManualAsset('Asset 2', 200.0, asset_class_map)))
         .add_account(
             Account('Account 2', 'Roth IRA')
             .add_asset(ManualAsset('Asset 1', 300.0, asset_class_map))
             .add_asset(ManualAsset('Asset 2', 400.0, asset_class_map))))

        self.assertAlmostEqual(1000.0, portfolio.total_value())
        self.assertEqual(
            'Account 1',
            portfolio.get_account('Account 1').name())
        self.assertEqual(
            'Account 2',
            portfolio.get_account('Account 2').name())
        self.assertAlmostEqual(
            100.0,
            portfolio.get_account('Account 1').get_asset('Asset 1').value())
        self.assertAlmostEqual(
            200.0,
            portfolio.get_account('Account 1').get_asset('Asset 2').value())
        self.assertAlmostEqual(
            300.0,
            portfolio.get_account('Account 2').get_asset('Asset 1').value())
        self.assertAlmostEqual(
            400.0,
            portfolio.get_account('Account 2').get_asset('Asset 2').value())
        self.assertListEqual(
            [['Account 1', 'Asset 1', '$100.00'],
             ['Account 1', 'Asset 2', '$200.00'],
             ['Account 2', 'Asset 1', '$300.00'],
             ['Account 2', 'Asset 2', '$400.00']],
            portfolio.assets().str_list())

    def test_get_account_name_by_substr(self):
        portfolio = Portfolio(AssetClass('All'))
        asset_class_map = {'All': 1.0}
        (portfolio
         .add_account(
             Account('Account 1', 'Taxable')
             .add_asset(ManualAsset('Asset 1', 100.0, asset_class_map))
             .add_asset(ManualAsset('Asset 2', 200.0, asset_class_map)))
         .add_account(
             Account('Account 2', 'Roth IRA')
             .add_asset(ManualAsset('Asset 1', 300.0, asset_class_map))
             .add_asset(ManualAsset('Asset 2', 400.0, asset_class_map))))
        self.assertEqual(
            'Account 1',
            portfolio.get_account_name_by_substr('1'))
        with self.assertRaisesRegex(AssertionError, 'matches more than'):
            portfolio.get_account_name_by_substr('Acc')
        with self.assertRaisesRegex(AssertionError, 'does not match'):
            portfolio.get_account_name_by_substr('God')

    def test_get_asset_name_by_substr(self):
        portfolio = Portfolio(AssetClass('All'))
        asset_class_map = {'All': 1.0}
        (portfolio
         .add_account(
             Account('Account 1', 'Taxable')
             .add_asset(ManualAsset('Asset 1', 100.0, asset_class_map))
             .add_asset(ManualAsset('Asset 2', 200.0, asset_class_map)))
         .add_account(
             Account('Account 2', 'Roth IRA')
             .add_asset(ManualAsset('Asset 1', 300.0, asset_class_map))
             .add_asset(ManualAsset('Funky Asset', 400.0, asset_class_map))))
        self.assertTupleEqual(('Account 2', 'Asset 1'),
                              portfolio.get_asset_name_by_substr(
            account_str='2', asset_str='1'))
        self.assertAlmostEqual(('Account 1', 'Asset 2'),
                               portfolio.get_asset_name_by_substr(
            account_str='Account', asset_str='2'))
        self.assertAlmostEqual(('Account 1', 'Asset 1'),
                               portfolio.get_asset_name_by_substr(
            account_str='1', asset_str='Asset 1'))
        self.assertAlmostEqual(('Account 2', 'Funky Asset'),
                               portfolio.get_asset_name_by_substr(
            asset_str='Funky'))

        with self.assertRaisesRegex(AssertionError, 'more than one'):
            portfolio.get_asset_name_by_substr(
                account_str='Acc', asset_str='Ass')
        with self.assertRaisesRegex(AssertionError, 'match none of'):
            portfolio.get_asset_name_by_substr(
                account_str='1', asset_str='Funky')
        with self.assertRaisesRegex(AssertionError, 'match none of'):
            portfolio.get_asset_name_by_substr(
                account_str='Acc', asset_str='Yolo')

    def test_what_ifs(self):
        portfolio = Portfolio(AssetClass('All')
                              .add_subclass(0.6, AssetClass('Equity'))
                              .add_subclass(0.4, AssetClass('Bonds')))
        asset1 = ManualAsset('Asset 1', 100.0, {'Equity': 1.0})
        asset2 = ManualAsset('Asset 2', 100.0, {'Bonds': 1.0})
        account1 = Account('Account 1', 'Taxable')
        account1.add_asset(asset1).add_asset(asset2)
        account2 = Account('Account 2', 'Pre-tax')
        portfolio.add_account(account1).add_account(account2)

        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual([], account_whatifs.list())
        self.assertListEqual([], asset_whatifs.list())

        portfolio.what_if('Account 1', 'Asset 2', -20)
        self.assertAlmostEqual(80, asset2.adjusted_value())
        self.assertAlmostEqual(20, account1.available_cash())
        self.assertAlmostEqual(200, portfolio.total_value())
        self.assertAlmostEqual(200, portfolio.total_value(False))
        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual([['Account 1', '+$20.00']],
                             account_whatifs.str_list())
        self.assertListEqual([['Account 1', 'Asset 2', '-$20.00']],
                             asset_whatifs.str_list())

        portfolio.what_if('Account 1', 'Asset 1', 20)
        self.assertAlmostEqual(120, asset1.adjusted_value())
        self.assertAlmostEqual(0, account1.available_cash())
        self.assertAlmostEqual(200, portfolio.total_value())
        self.assertAlmostEqual(200, portfolio.total_value(False))
        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual([], account_whatifs.str_list())
        self.assertListEqual(
            [['Account 1', 'Asset 1', '+$20.00'],
             ['Account 1', 'Asset 2', '-$20.00']],
            asset_whatifs.str_list())

        self.assertListEqual(
            [['Account 1', 'Asset 1', '$120.00'],
             ['Account 1', 'Asset 2', '$80.00']],
            portfolio.assets().str_list())

        self.assertListEqual(
            [['All:'],
             ['Equity', '60.0%', '60.0%', '$120.00'],
             ['Bonds', '40.0%', '40.0%', '$80.00']],
            portfolio.asset_allocation_tree().str_list())

        portfolio.what_if_add_cash('Account 1', 30)
        self.assertAlmostEqual(30, account1.available_cash())
        self.assertAlmostEqual(230, portfolio.total_value())
        self.assertAlmostEqual(200, portfolio.total_value(False))
        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual(
            [['Account 1', '+$30.00']],
            account_whatifs.str_list())
        self.assertListEqual(
            [['Account 1', 'Asset 1', '+$20.00'],
             ['Account 1', 'Asset 2', '-$20.00']],
            asset_whatifs.str_list())

        portfolio.what_if_add_cash('Account 2', 460)
        self.assertAlmostEqual(460, account2.available_cash())
        self.assertAlmostEqual(690, portfolio.total_value())
        self.assertAlmostEqual(200, portfolio.total_value(False))
        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual(
            [['Account 1', '+$30.00'],
             ['Account 2', '+$460.00']],
            account_whatifs.str_list())
        self.assertListEqual(
            [['Account 1', 'Asset 1', '+$20.00'],
             ['Account 1', 'Asset 2', '-$20.00']],
            asset_whatifs.str_list())

        self.assertListEqual(
            [['Equity', 'Taxable', '100.0%', '$120.00'],
             ['Bonds', 'Taxable', '100.0%', '$80.00']],
            portfolio.asset_location().str_list())

        portfolio.reset_what_ifs()
        self.assertAlmostEqual(100, asset1.adjusted_value())
        self.assertAlmostEqual(100, asset2.adjusted_value())
        self.assertAlmostEqual(0, account1.available_cash())
        self.assertAlmostEqual(0, account2.available_cash())
        self.assertAlmostEqual(200, portfolio.total_value())
        self.assertAlmostEqual(200, portfolio.total_value(False))
        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual([], account_whatifs.str_list())
        self.assertListEqual([], asset_whatifs.str_list())

    @patch('lakshmi.assets.TickerAsset.name')
    @patch('lakshmi.assets.TickerAsset.price')
    def test_get_what_ifs_options(self, mock_price, mock_name):
        mock_price.return_value = 2.0
        mock_name.return_value = 'Vanguard Cash Reserves Federal'

        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable')
            .add_asset(TickerAsset('VMMXX', 420.0, {'All': 1.0}))
            .add_asset(ManualAsset('Cash', 840.0, {'All': 1.0})))

        portfolio.what_if('Schwab', 'VMMXX', -20)
        portfolio.what_if('Schwab', 'Cash', 20)
        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual(
            [['Schwab', 'Vanguard Cash Reserves Federal', '-$20.00'],
             ['Schwab', 'Cash', '+$20.00']],
            asset_whatifs.str_list())
        account_whatifs, asset_whatifs = portfolio.get_what_ifs(
            long_name=False, short_name=True)
        self.assertListEqual(
            [['Schwab', 'VMMXX', '-$20.00'],
             ['Schwab', 'Cash', '+$20.00']],
            asset_whatifs.str_list())
        account_whatifs, asset_whatifs = portfolio.get_what_ifs(
            long_name=False, short_name=True, quantity=True)
        self.assertListEqual(
            [['Schwab', 'VMMXX', '-10', '-$20.00'],
             ['Schwab', 'Cash', '', '+$20.00']],
            asset_whatifs.str_list())

    def test_what_ifs_double_add(self):
        portfolio = Portfolio(AssetClass('All'))

        asset = ManualAsset('Asset', 100.0, {'All': 1.0})
        account = Account('Account', 'Taxable')
        portfolio.add_account(account.add_asset(asset))

        portfolio.what_if('Account', 'Asset', 20)
        portfolio.what_if('Account', 'Asset', 30)
        self.assertAlmostEqual(150, asset.adjusted_value())
        self.assertAlmostEqual(-50, account.available_cash())
        self.assertAlmostEqual(100, portfolio.total_value())
        account_whatifs, asset_whatifs = portfolio.get_what_ifs()
        self.assertListEqual(
            [['Account', '-$50.00']],
            account_whatifs.str_list())
        self.assertListEqual(
            [['Account', 'Asset', '+$50.00']],
            asset_whatifs.str_list())

    def test_return_allocation_one_asset(self):
        asset_class = AssetClass('All').validate()
        allocation = {'All': 10.0}

        ret = asset_class.return_allocation(allocation)
        self.assertEqual(1, len(ret))
        self.assertEqual('All', ret[0].name)
        self.assertAlmostEqual(10.0, ret[0].value)
        self.assertEqual([], ret[0].children)

    def test_return_allocation_asset_tree(self):
        asset_class = (
            AssetClass('All')
            .add_subclass(0.6, AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.4, AssetClass('Bonds'))).validate()
        allocation = {'US': 10.0, 'Intl': 20.0, 'Bonds': 40.0}

        ret = asset_class.return_allocation(allocation, levels=0)

        self.assertEqual(1, len(ret))
        self.assertEqual('All', ret[0].name)
        self.assertAlmostEqual(70.0, ret[0].value)

        self.assertEqual(2, len(ret[0].children))
        self.assertEqual('Equity', ret[0].children[0].name)
        self.assertAlmostEqual(
            0.429, ret[0].children[0].actual_allocation, places=3)
        self.assertAlmostEqual(0.6, ret[0].children[0].desired_allocation)
        self.assertAlmostEqual(12.0, ret[0].children[0].value_difference)
        self.assertEqual('Bonds', ret[0].children[1].name)
        self.assertAlmostEqual(
            0.571, ret[0].children[1].actual_allocation, places=3)
        self.assertAlmostEqual(0.4, ret[0].children[1].desired_allocation)
        self.assertAlmostEqual(-12.0, ret[0].children[1].value_difference)

        self.assertEqual(
            3, len(
                asset_class.return_allocation(
                    allocation, levels=1)))
        self.assertEqual(
            5, len(
                asset_class.return_allocation(
                    allocation, levels=2)))
        self.assertEqual(5, len(asset_class.return_allocation(allocation)))
        self.assertEqual(
            5, len(
                asset_class.return_allocation(
                    allocation, levels=5)))

    @patch('lakshmi.assets.TickerAsset.name')
    @patch('lakshmi.assets.TickerAsset.price')
    def test_list_lots(self, mock_price, mock_name):
        mock_price.return_value = 200.0
        mock_name.return_value = 'Unused'

        vti = TickerAsset('VTI', 100.0, {'All': 1.0})
        vti.set_lots([TaxLot('2020/01/01', 50, 100.0),
                      TaxLot('2021/01/01', 50, 300.0)])
        vxus = TickerAsset('VXUS', 50.0, {'All': 1.0})
        vxus.set_lots([TaxLot('2019/01/01', 50, 150.0)])
        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable')
            .add_asset(vti)
            .add_asset(ManualAsset('Cash', 840.0, {'All': 1.0}))
            .add_asset(vxus))
        # Order of lots: ShortName, Date, Cost, Gain, Gain%
        self.assertListEqual(
            [['VTI', '2020/01/01', '$5,000.00', '+$5,000.00', '100.0%'],
             ['VTI', '2021/01/01', '$15,000.00', '-$5,000.00', '-33.3%'],
             ['VXUS', '2019/01/01', '$7,500.00', '+$2,500.00', '33.3%']],
            portfolio.list_lots().str_list())

    @patch('lakshmi.assets.TickerAsset.name')
    @patch('lakshmi.assets.TickerAsset.price')
    def test_list_lots_with_account_and_term(self, mock_price, mock_name):
        mock_price.return_value = 200.0
        mock_name.return_value = 'Unused'

        vti = TickerAsset('VTI', 100.0, {'All': 1.0})
        vti.set_lots([TaxLot('2020/01/01', 50, 100.0),
                      TaxLot('2021/01/01', 50, 300.0)])
        vxus = TickerAsset('VXUS', 50.0, {'All': 1.0})
        vxus.set_lots([TaxLot('2019/01/01', 50, 150.0)])
        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable')
            .add_asset(vti)
            .add_asset(ManualAsset('Cash', 840.0, {'All': 1.0}))
            .add_asset(vxus))
        # Order of lots: Account, ShortName, Date, Cost, Gain, Gain%
        self.assertListEqual(
            [['Schwab', 'VTI', '2020/01/01', '$5,000.00', '+$5,000.00',
              '100.0%'],
             ['Schwab', 'VTI', '2021/01/01', '$15,000.00', '-$5,000.00',
              '-33.3%'],
             ['Schwab', 'VXUS', '2019/01/01', '$7,500.00', '+$2,500.00',
              '33.3%']],
            portfolio.list_lots(include_account=True).str_list())

        # We just check that term is included in the row instead of exact
        # calculatin of term
        self.assertListEqual(
            [6, 6, 6],
            list(map(len, portfolio.list_lots(include_term=True).list())))
        self.assertListEqual(
            [7, 7, 7],
            list(map(len, portfolio.list_lots(
                include_account=True, include_term=True).list())))


if __name__ == '__main__':
    unittest.main()



"""Tests for lakshmi.table module."""
import unittest

from lakshmi.table import Table


class TableTest(unittest.TestCase):
    def test_sanity(self):
        self.assertEqual(len(Table.coltype2func), len(Table.coltype2align))

    def test_empty_table(self):
        t = Table(3)
        self.assertListEqual([], t.list())
        self.assertListEqual([], t.str_list())
        self.assertEqual('', t.string())

    def test_no_headers_and_coltypes(self):
        t = Table(3)
        t.add_row(['1', '2', '3'])
        self.assertListEqual([['1', '2', '3']], t.list())
        self.assertListEqual([['1', '2', '3']], t.str_list())
        self.assertGreater(len(t.string()), 0)

    def test_bad_coltypes(self):
        with self.assertRaisesRegex(AssertionError,
                                    'Bad column type in coltypes'):
            Table(2, coltypes=[None, 'str'])

    def test_set_rows(self):
        t = Table(3)
        t.set_rows([['1', '2']])
        self.assertListEqual([['1', '2']], t.str_list())

    def test_headers_and_diff_coltypes(self):
        headers = ['1', '2', '3', '4', '5', '6']
        t = Table(
            6,
            headers=headers,
            coltypes=[
                'str',
                'dollars',
                'delta_dollars',
                'percentage',
                'percentage_1',
                'float'])

        rows = [['r1', 3, 4.1, 0.5, 0.5, 1],
                ['r6', 8, -9.2, 0.1, 0.5557, 2.345]]
        t.set_rows(rows)

        self.assertListEqual(headers, t.headers())
        self.assertListEqual(
            ['left', 'right', 'right', 'right', 'right', 'decimal'],
            t.col_align())

        self.assertListEqual(rows, t.list())
        self.assertListEqual(
            [['r1', '$3.00', '+$4.10', '50%', '50.0%', '1.0'],
             ['r6', '$8.00', '-$9.20', '10%', '55.6%', '2.345']],
            t.str_list())
        self.assertGreater(len(t.string()), 0)

    def test_mismatched_num_cols(self):
        with self.assertRaises(AssertionError):
            Table(2, headers=['1'])
        with self.assertRaises(AssertionError):
            Table(2, headers=['1', '2', '3'])
        with self.assertRaises(AssertionError):
            Table(2, coltypes=['str'])
        with self.assertRaises(AssertionError):
            Table(2, headers=['str', 'str', 'str'])

    def test_too_many_cols(self):
        t = Table(2)
        with self.assertRaises(AssertionError):
            t.add_row(['1', '2', '3'])
        with self.assertRaises(AssertionError):
            t.set_rows([['a', 'b'], ['1', '2', '3']])

    def test_too_few_cols(self):
        t = Table(2)
        t.add_row(['1'])
        t.set_rows([['1', '2'], ['1'], ['a']])


if __name__ == '__main__':
    unittest.main()



"""Tests for lakshmi.assets module."""
import unittest
from unittest.mock import patch

import lakshmi.analyze as analyze
from lakshmi import Account, AssetClass, Portfolio
from lakshmi.assets import ManualAsset, TaxLot, TickerAsset


class AnalyzeTest(unittest.TestCase):
    def test_tlh_no_lots(self):
        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable').add_asset(
                ManualAsset('Cash', 100.0, {'All': 1.0})))
        self.assertListEqual(
            [],
            analyze.TLH(0.5, 10000).analyze(portfolio).list())

    @patch('lakshmi.assets.TickerAsset.price')
    def test_tlh_loss_percent(self, mock_price):
        mock_price.return_value = 10.0
        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable')
            .add_asset(TickerAsset('VTI', 1000.0, {'All': 1.0}).set_lots(
                [TaxLot('2020/01/31', 500.0, 21.0),
                 TaxLot('2021/03/31', 250.0, 14.9),
                 TaxLot('2021/05/31', 250.0, 8.0)]))
            .add_asset(TickerAsset('VXUS', 100.0, {'All': 1.0}).set_lots(
                [TaxLot('2020/01/25', 100.0, 30.0)]))
            .add_asset(TickerAsset('ITOT', 200.0, {'All': 1.0}).set_lots(
                [TaxLot('2019/09/01', 200.0, 12.0)]))
            .add_asset(TickerAsset('VOO', 500.0, {'All': 1.0})))
        self.assertListEqual(
            [['Schwab', 'VTI', '2020/01/31', '$5,500.00', '52%'],
             ['Schwab', 'VXUS', '2020/01/25', '$2,000.00', '67%']],
            analyze.TLH(0.5).analyze(portfolio).str_list())

    @patch('lakshmi.assets.TickerAsset.price')
    def test_tlh_combo(self, mock_price):
        mock_price.return_value = 10.0
        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable')
            .add_asset(TickerAsset('VTI', 1000.0, {'All': 1.0}).set_lots(
                [TaxLot('2020/01/31', 500.0, 11.0),
                 TaxLot('2021/03/31', 500.0, 12.0)]))
            .add_asset(TickerAsset('VXUS', 100.0, {'All': 1.0}).set_lots(
                [TaxLot('2020/01/25', 100.0, 30.0)]))
            .add_asset(TickerAsset('ITOT', 200.0, {'All': 1.0}).set_lots(
                [TaxLot('2019/09/01', 200.0, 12.0)]))
            .add_asset(TickerAsset('VOO', 500.0, {'All': 1.0})))
        self.assertListEqual(
            [['Schwab', 'VTI', '2020/01/31', '$500.00', '9%'],
             ['Schwab', 'VTI', '2021/03/31', '$1,000.00', '17%'],
             ['Schwab', 'VXUS', '2020/01/25', '$2,000.00', '67%']],
            analyze.TLH(0.5, 1400).analyze(portfolio).str_list())

    def test_rebalance_outside_bounds(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 56.0, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 30.0, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 14.0, {'Bond': 1.0})))

        self.assertListEqual(
            [['Bond', '14%', '10%', '$14.00', '-$4.00'],
             ['Intl', '30%', '36%', '$30.00', '+$6.00']],
            sorted(analyze.BandRebalance().analyze(portfolio).str_list()))

    def test_rebalance_within_bounds(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 54.0, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 35.0, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 11.0, {'Bond': 1.0})))

        self.assertListEqual(
            [],
            analyze.BandRebalance().analyze(portfolio).list())

    def test_allocate_no_cash(self):
        portfolio = Portfolio(AssetClass('All')).add_account(
            Account('Schwab', 'Taxable').add_asset(
                ManualAsset('Cash', 100.0, {'All': 1.0})))
        with self.assertRaisesRegex(AssertionError, 'No available cash'):
            analyze.Allocate('Schwab').analyze(portfolio)

    def test_allocate_not_enough_assets(self):
        account = Account('Schwab', 'Taxable').add_asset(
            ManualAsset('Cash', 100.0, {'All': 1.0}))
        account.add_cash(-200)
        portfolio = Portfolio(AssetClass('All')).add_account(account)
        with self.assertRaisesRegex(AssertionError, 'Cash to withdraw'):
            analyze.Allocate('Schwab').analyze(portfolio)

    def test_allocate_not_enough_assets_blacklisted(self):
        account = (
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Cash', 100.0, {'All': 1.0}))
            .add_asset(ManualAsset('Black Cash', 100.0, {'All': 1.0})))
        account.add_cash(-150)
        portfolio = Portfolio(AssetClass('All')).add_account(account)
        with self.assertRaisesRegex(AssertionError, 'Cash to withdraw'):
            analyze.Allocate(
                'Schwab', exclude_assets=['Black Cash']).analyze(portfolio)

    def test_allocate_no_assets(self):
        account = Account('Schwab', 'Taxable').add_asset(
            ManualAsset('Cash', 100.0, {'All': 1.0}))
        account.add_cash(200)
        portfolio = Portfolio(AssetClass('All')).add_account(account)
        with self.assertRaisesRegex(AssertionError, 'No assets to allocate'):
            analyze.Allocate('Schwab', ['Cash']).analyze(portfolio)

    def test_allocate_one_assets(self):
        account = Account('Schwab', 'Taxable').add_asset(
            ManualAsset('Cash', 100.0, {'All': 1.0}))
        account.add_cash(200)
        portfolio = Portfolio(AssetClass('All')).add_account(account)
        self.assertListEqual(
            [['Cash', '+$200.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())

    def test_allocate_cash(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 53.0, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 35.0, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 9.0, {'Bond': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(3)
        self.assertListEqual(
            [['Total US', '+$1.00'],
             ['Total Intl', '+$1.00'],
             ['Total Bond', '+$1.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())
        self.assertAlmostEqual(
            54, account.get_asset('Total US').adjusted_value(), places=6)
        self.assertAlmostEqual(
            36, account.get_asset('Total Intl').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

        account.add_cash(100)
        self.assertListEqual(
            [['Total US', '+$54.00'],
             ['Total Intl', '+$36.00'],
             ['Total Bond', '+$10.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())
        self.assertAlmostEqual(
            54 * 2, account.get_asset('Total US').adjusted_value(), places=6)
        self.assertAlmostEqual(
            36 * 2, account.get_asset('Total Intl').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10 * 2, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

        account.add_cash(-200)
        self.assertListEqual(
            [['Total US', '-$108.00'],
             ['Total Intl', '-$72.00'],
             ['Total Bond', '-$20.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())
        self.assertAlmostEqual(
            0, account.get_asset('Total US').adjusted_value(), places=6)
        self.assertAlmostEqual(
            0, account.get_asset('Total Intl').adjusted_value(), places=6)
        self.assertAlmostEqual(
            0, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_cash_indentical_assets(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 53.0, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 17.5, {'Intl': 1.0}))
            .add_asset(ManualAsset('Ex US', 17.5, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 9.0, {'Bond': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(3)
        self.assertListEqual(
            [['Total US', '+$1.00'],
             ['Total Intl', '+$0.50'],
             ['Ex US', '+$0.50'],
             ['Total Bond', '+$1.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())

    def test_allocate_cash_overlapping_assets(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4,
                                        AssetClass('Intl')
                                        .add_subclass(
                                            0.7, AssetClass('Developed'))
                                        .add_subclass(
                                            0.3, AssetClass('Emerging'))))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 53.0, {'US': 1.0}))
            .add_asset(ManualAsset('Intl', 17.5, {'Developed': 0.7,
                                                  'Emerging': 0.3}))
            .add_asset(ManualAsset('Devel', 12.25, {'Developed': 1.0}))
            .add_asset(ManualAsset('Emer', 5.25, {'Emerging': 1.0}))
            .add_asset(ManualAsset('Total Bond', 9.0, {'Bond': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(3)
        self.assertListEqual(
            [['Total US', '+$1.00'],
             ['Intl', '+$0.33'],
             ['Devel', '+$0.47'],
             ['Emer', '+$0.20'],
             ['Total Bond', '+$1.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())

    def test_allocate_only_rebalance(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 60.0, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 40.0, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 0, {'Bond': 1.0})))

        account = portfolio.get_account('Schwab')
        self.assertListEqual(
            [['Total US', '-$6.00'],
             ['Total Intl', '-$4.00'],
             ['Total Bond', '+$10.00']],
            analyze.Allocate('Schwab', rebalance=True).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            54, account.get_asset('Total US').adjusted_value(), places=6)
        self.assertAlmostEqual(
            36, account.get_asset('Total Intl').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_cash_rebalance(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 50.0, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 40.0, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 0, {'Bond': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(10)
        self.assertListEqual(
            [['Total US', '+$4.00'],
             ['Total Intl', '-$4.00'],
             ['Total Bond', '+$10.00']],
            analyze.Allocate('Schwab', rebalance=True).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            54, account.get_asset('Total US').adjusted_value(), places=6)
        self.assertAlmostEqual(
            36, account.get_asset('Total Intl').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_cash_large_portfolio(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 53e6, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 35e6, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 9e6, {'Bond': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(3e6)
        self.assertListEqual(
            [['Total US', '+$1,000,000.00'],
             ['Total Intl', '+$1,000,000.00'],
             ['Total Bond', '+$1,000,000.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())
        self.assertAlmostEqual(
            54e6, account.get_asset('Total US').adjusted_value(), places=6)
        self.assertAlmostEqual(
            36e6, account.get_asset('Total Intl').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10e6, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_cash_rebalance_large_portfolio(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4, AssetClass('Intl')))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 50e6, {'US': 1.0}))
            .add_asset(ManualAsset('Total Intl', 40e6, {'Intl': 1.0}))
            .add_asset(ManualAsset('Total Bond', 0, {'Bond': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(10e6)
        self.assertListEqual(
            [['Total US', '+$4,000,000.00'],
             ['Total Intl', '-$4,000,000.00'],
             ['Total Bond', '+$10,000,000.00']],
            analyze.Allocate('Schwab', rebalance=True).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            54e6, account.get_asset('Total US').adjusted_value(), places=6)
        self.assertAlmostEqual(
            36e6, account.get_asset('Total Intl').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10e6, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_cash_going_negative_corr(self):
        # In this test case, due to coorelation between funds, the solution
        # will drive Developed/emerging funds to negative balance.
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4,
                                        AssetClass('Intl')
                                        .add_subclass(
                                            0.7, AssetClass('Developed'))
                                        .add_subclass(
                                            0.3, AssetClass('Emerging'))))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 54.0, {'US': 1.0}))
            .add_asset(ManualAsset('Intl', 16.2, {'Developed': 0.7,
                                                  'Emerging': 0.3}))
            .add_asset(ManualAsset('Devel', 11.34, {'Developed': 1.0}))
            .add_asset(ManualAsset('Emer', 4.86, {'Emerging': 1.0}))
            .add_asset(ManualAsset('Total Bond', 10.0, {'Bond': 1.0})))

        portfolio.get_account('Schwab').add_cash(3)
        self.assertListEqual(
            [['Total US', '+$0.00'],
             ['Intl', '+$0.98'],
             ['Devel', '+$1.36'],
             ['Emer', '+$0.66'],
             ['Total Bond', '+$0.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())

    def test_allocate_cash_going_positive_corr(self):
        # In this test case, due to coorelation between funds, the solution
        # will drive Developed/emerging funds to positive balance.
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9,
                          AssetClass('Equity')
                          .add_subclass(0.6, AssetClass('US'))
                          .add_subclass(0.4,
                                        AssetClass('Intl')
                                        .add_subclass(
                                            0.7, AssetClass('Developed'))
                                        .add_subclass(
                                            0.3, AssetClass('Emerging'))))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total US', 54.0, {'US': 1.0}))
            .add_asset(ManualAsset('Intl', 19.8, {'Developed': 0.7,
                                                  'Emerging': 0.3}))
            .add_asset(ManualAsset('Devel', 13.86, {'Developed': 1.0}))
            .add_asset(ManualAsset('Emer', 5.94, {'Emerging': 1.0}))
            .add_asset(ManualAsset('Total Bond', 10.0, {'Bond': 1.0})))

        portfolio.get_account('Schwab').add_cash(-3)
        self.assertListEqual(
            [['Total US', '+$0.00'],
             ['Intl', '-$0.98'],
             ['Devel', '-$1.36'],
             ['Emer', '-$0.66'],
             ['Total Bond', '+$0.00']],
            analyze.Allocate('Schwab').analyze(portfolio).str_list())

    def test_allocate_cash_withdrawing_more(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9, AssetClass('Equity'))
            .add_subclass(0.1, AssetClass('Bond')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('US', 90, {'Equity': 1.0}))
            .add_asset(ManualAsset('Bonds', 20, {'Bond': 1.0})))
        portfolio.add_account(
            Account('401K', 'Tax-Deferred')
            .add_asset(ManualAsset('US', 3, {'Equity': 1.0}))
            .add_asset(ManualAsset('Bonds', 1, {'Bond': 1.0})))

        portfolio.get_account('401K').add_cash(-2)

        # We should ideally withdraw money from bonds, but it doesn't have
        # any balance to withdraw.
        self.assertListEqual(
            [['US', '-$1.00'],
             ['Bonds', '-$1.00']],
            analyze.Allocate('401K').analyze(portfolio).str_list())

    def test_allocate_zero_ratio(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9, AssetClass('Equity'))
            .add_subclass(0.1, AssetClass('Bond'))
            .add_subclass(0, AssetClass('Zero')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total Market', 90.0, {'Equity': 1.0}))
            .add_asset(ManualAsset('Total Bond', 10, {'Bond': 1.0}))
            .add_asset(ManualAsset('Total Zero', 10, {'Zero': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(10)
        # These numbers "look" incorrect, but one can manually verify that the
        # defined error function is minimized at this point.
        self.assertListEqual(
            [['Total Market', '+$8.12'],
             ['Total Bond', '+$1.88'],
             ['Total Zero', '+$0.00']],
            analyze.Allocate('Schwab', rebalance=False).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            98.12, account.get_asset('Total Market').adjusted_value(),
            places=2)
        self.assertAlmostEqual(
            11.88, account.get_asset('Total Bond').adjusted_value(), places=2)
        self.assertAlmostEqual(
            10, account.get_asset('Total Zero').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_zero_ratio_withdraw(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9, AssetClass('Equity'))
            .add_subclass(0.1, AssetClass('Bond'))
            .add_subclass(0, AssetClass('Zero')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total Market', 90.0, {'Equity': 1.0}))
            .add_asset(ManualAsset('Total Bond', 10, {'Bond': 1.0}))
            .add_asset(ManualAsset('Total Zero', 10, {'Zero': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(-10)
        self.assertListEqual(
            [['Total Market', '+$0.00'],
             ['Total Bond', '+$0.00'],
             ['Total Zero', '-$10.00']],
            analyze.Allocate('Schwab', rebalance=False).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            90, account.get_asset('Total Market').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(
            0, account.get_asset('Total Zero').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_zero_ratio_withdraw_less(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9, AssetClass('Equity'))
            .add_subclass(0.1, AssetClass('Bond'))
            .add_subclass(0, AssetClass('Zero')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total Market', 90.0, {'Equity': 1.0}))
            .add_asset(ManualAsset('Total Bond', 10, {'Bond': 1.0}))
            .add_asset(ManualAsset('Total Zero', 10, {'Zero': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(-5)
        self.assertListEqual(
            [['Total Market', '+$0.00'],
             ['Total Bond', '+$0.00'],
             ['Total Zero', '-$5.00']],
            analyze.Allocate('Schwab', rebalance=False).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            90, account.get_asset('Total Market').adjusted_value(), places=6)
        self.assertAlmostEqual(
            10, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(
            5, account.get_asset('Total Zero').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_zero_ratio_withdraw_more(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9, AssetClass('Equity'))
            .add_subclass(0.1, AssetClass('Bond'))
            .add_subclass(0, AssetClass('Zero')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total Market', 90.0, {'Equity': 1.0}))
            .add_asset(ManualAsset('Total Bond', 10, {'Bond': 1.0}))
            .add_asset(ManualAsset('Total Zero', 10, {'Zero': 1.0})))

        account = portfolio.get_account('Schwab')
        account.add_cash(-20)
        self.assertListEqual(
            [['Total Market', '-$9.00'],
             ['Total Bond', '-$1.00'],
             ['Total Zero', '-$10.00']],
            analyze.Allocate('Schwab', rebalance=False).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            81, account.get_asset('Total Market').adjusted_value(), places=6)
        self.assertAlmostEqual(
            9, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(
            0, account.get_asset('Total Zero').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)

    def test_allocate_zero_ratio_rebalance(self):
        portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.9, AssetClass('Equity'))
            .add_subclass(0.1, AssetClass('Bond'))
            .add_subclass(0, AssetClass('Zero')).validate())
        portfolio.add_account(
            Account('Schwab', 'Taxable')
            .add_asset(ManualAsset('Total Market', 90.0, {'Equity': 1.0}))
            .add_asset(ManualAsset('Total Bond', 10, {'Bond': 1.0}))
            .add_asset(ManualAsset('Total Zero', 10, {'Zero': 1.0})))

        account = portfolio.get_account('Schwab')
        self.assertListEqual(
            [['Total Market', '+$9.00'],
             ['Total Bond', '+$1.00'],
             ['Total Zero', '-$10.00']],
            analyze.Allocate('Schwab', rebalance=True).analyze(
                portfolio).str_list())
        self.assertAlmostEqual(
            99, account.get_asset('Total Market').adjusted_value(), places=6)
        self.assertAlmostEqual(
            11, account.get_asset('Total Bond').adjusted_value(), places=6)
        self.assertAlmostEqual(
            0, account.get_asset('Total Zero').adjusted_value(), places=6)
        self.assertAlmostEqual(0, account.available_cash(), places=6)


if __name__ == '__main__':
    unittest.main()



"""Test for lakshmi.performance module."""

import unittest
from datetime import datetime

from lakshmi.performance import Checkpoint, Performance, Timeline


class PerformanceTest(unittest.TestCase):

    def test_checkpoint(self):
        c = Checkpoint('2020/11/11', 100)  # Default constructor.
        self.assertEqual(100, c.get_portfolio_value())
        self.assertEqual('2020/11/11', c.get_date())

        c = Checkpoint.from_dict(c.to_dict())
        self.assertEqual(100, c.get_portfolio_value())
        self.assertEqual('2020/11/11', c.get_date())

        c = Checkpoint('2020/1/1', 200, inflow=100, outflow=50)
        self.assertEqual(200, c.get_portfolio_value())
        self.assertEqual('2020/01/01', c.get_date())
        self.assertEqual(100, c.get_inflow())
        self.assertEqual(50, c.get_outflow())

        c = Checkpoint.from_dict(c.to_dict())
        self.assertEqual(200, c.get_portfolio_value())
        self.assertEqual('2020/01/01', c.get_date())
        self.assertEqual(100, c.get_inflow())
        self.assertEqual(50, c.get_outflow())

        self.assertNotIn('Date', c.to_dict(show_date=False))

        c = Checkpoint.from_dict(c.to_dict(show_date=False), '2020/12/12')
        self.assertEqual(200, c.get_portfolio_value())
        self.assertEqual('2020/12/12', c.get_date())
        self.assertEqual(100, c.get_inflow())
        self.assertEqual(50, c.get_outflow())

    def test_show_empty(self):
        c = Checkpoint('2020/11/11', 100, inflow=10)
        d = c.to_dict()
        self.assertFalse('Outflow' in d)
        d = c.to_dict(show_empty_cashflow=True)
        self.assertTrue('Outflow' in d)

    def test_empty_timeline(self):
        with self.assertRaises(AssertionError):
            Timeline([])

    def test_single_entry_timeline(self):
        cp = Checkpoint('2021/1/1', 100.0)
        timeline = Timeline([cp])

        self.assertTrue(timeline.has_checkpoint('2021/01/1'))
        self.assertFalse(timeline.has_checkpoint('2021/01/02'))

        self.assertEqual('2021/01/01', timeline.begin())
        self.assertEqual('2021/01/01', timeline.end())

        self.assertFalse(timeline.covers('2020/12/31'))
        self.assertTrue(timeline.covers('2021/1/1'))

        self.assertEqual(100, timeline.get_checkpoint('2021/01/1')
                         .get_portfolio_value())
        self.assertEqual(100, timeline.get_checkpoint('2021/01/1', True)
                         .get_portfolio_value())
        with self.assertRaises(AssertionError):
            timeline.get_checkpoint('2021/01/02')
        with self.assertRaises(AssertionError):
            timeline.get_checkpoint('2021/01/02', True)

        with self.assertRaises(AssertionError):
            timeline.insert_checkpoint(cp)

        cp_replace = Checkpoint('2021/1/1', 150.0)
        timeline.insert_checkpoint(cp_replace, replace=True)
        self.assertEqual(150, timeline.get_checkpoint('2021/1/1')
                         .get_portfolio_value())

        cp1 = Checkpoint('2020/1/1', 200)
        timeline.insert_checkpoint(cp1)
        self.assertEqual('2020/01/01', timeline.begin())

    def test_timeline(self):
        checkpoints = [
            Checkpoint('2021/1/1', 100),
            Checkpoint('2021/3/1', 500),
            Checkpoint('2021/1/31', 300, inflow=150, outflow=50)]
        timeline = Timeline(checkpoints)

        self.assertTrue(timeline.has_checkpoint('2021/01/1'))
        self.assertFalse(timeline.has_checkpoint('2021/01/02'))

        self.assertEqual('2021/01/01', timeline.begin())
        self.assertEqual('2021/03/01', timeline.end())

        self.assertFalse(timeline.covers('2020/12/31'))
        self.assertTrue(timeline.covers('2021/1/1'))
        self.assertTrue(timeline.covers('2021/1/2'))
        self.assertTrue(timeline.covers('2021/3/1'))

        self.assertEqual(300, timeline.get_checkpoint('2021/01/31')
                         .get_portfolio_value())
        self.assertEqual(300, timeline.get_checkpoint('2021/01/31', True)
                         .get_portfolio_value())

        with self.assertRaises(AssertionError):
            timeline.get_checkpoint('2021/01/15')

        self.assertEqual(150, timeline.get_checkpoint('2021/01/16', True)
                         .get_portfolio_value())

    def test_timeline_to_list(self):
        checkpoints = [
            Checkpoint('2021/1/1', 100),
            Checkpoint('2021/3/1', 500),
            Checkpoint('2021/1/31', 300, inflow=150, outflow=50)]
        timeline = Timeline(checkpoints)
        timeline_list = timeline.to_list()
        self.assertEqual([
            {'Date': '2021/01/01', 'Portfolio Value': 100},
            {'Date': '2021/01/31', 'Portfolio Value': 300,
             'Inflow': 150, 'Outflow': 50},
            {'Date': '2021/03/01', 'Portfolio Value': 500}],
            timeline_list)
        timeline = Timeline.from_list(timeline_list)
        self.assertTrue(timeline.has_checkpoint('2021/01/01'))
        self.assertTrue(timeline.has_checkpoint('2021/01/31'))
        self.assertTrue(timeline.has_checkpoint('2021/03/01'))

    def test_timeline_to_table(self):
        checkpoints = [
            Checkpoint('2021/1/1', 100),
            Checkpoint('2021/3/1', 500),
            Checkpoint('2021/1/31', 300, inflow=150, outflow=50)]
        timeline = Timeline(checkpoints)
        self.assertEqual(
            [['2021/01/01', '$100.00', '$0.00', '$0.00'],
             ['2021/01/31', '$300.00', '$150.00', '$50.00'],
             ['2021/03/01', '$500.00', '$0.00', '$0.00']],
            timeline.to_table().str_list())
        self.assertEqual(
            [['2021/01/01', '$100.00', '$0.00', '$0.00']],
            timeline.to_table('2021/01/01', '2021/01/02').str_list())
        self.assertEqual(
            [['2021/03/01', '$500.00', '$0.00', '$0.00']],
            timeline.to_table('2021/02/01', '2021/04/01').str_list())
        self.assertEqual(
            [['2021/01/31', '$300.00', '$150.00', '$50.00']],
            timeline.to_table('2021/01/31', '2021/01/31').str_list())
        self.assertEqual(
            [], timeline.to_table('2021/01/15', '2021/01/17').str_list())

    def test_get_performance_data(self):
        checkpoints = [
            Checkpoint('2021/1/1', 100),
            Checkpoint('2021/1/31', 300, inflow=150, outflow=50),
            Checkpoint('2021/3/1', 500, inflow=10, outflow=20)]
        timeline = Timeline(checkpoints)
        data = timeline.get_performance_data('2021/01/01', '2021/03/01')
        self.assertEqual(
            [datetime(2021, 1, 1), datetime(2021, 1, 31),
             datetime(2021, 3, 1)],
            data.dates)
        self.assertEqual([-100, -100, 510], data.amounts)
        self.assertEqual(100, data.begin_balance)
        self.assertEqual(500, data.end_balance)
        self.assertEqual(160, data.inflows)
        self.assertEqual(70, data.outflows)

        data = timeline.get_performance_data('2021/01/16', '2021/01/31')
        self.assertEqual([datetime(2021, 1, 16), datetime(2021, 1, 31)],
                         data.dates)
        self.assertEqual([-150, 200], data.amounts)
        self.assertEqual(150, data.begin_balance)
        self.assertEqual(300, data.end_balance)
        self.assertEqual(150, data.inflows)
        self.assertEqual(50, data.outflows)

    def test_get_performance_data_none_dates(self):
        checkpoints = [
            Checkpoint('2021/1/1', 100),
            Checkpoint('2021/1/31', 300, inflow=150, outflow=50),
            Checkpoint('2021/3/1', 500, inflow=10, outflow=20)]
        timeline = Timeline(checkpoints)
        data = timeline.get_performance_data(None, None)
        self.assertEqual(100, data.begin_balance)
        self.assertEqual(500, data.end_balance)

    def test_performance_to_dict(self):
        perf = Performance(Timeline([Checkpoint('2021/1/1', 100)]))
        perf = Performance.from_dict(perf.to_dict())
        self.assertEqual('2021/01/01', perf.get_timeline().begin())
        self.assertEqual('2021/01/01', perf.get_timeline().end())

    def test_summary_table_single_date(self):
        perf_table = Performance(Timeline([
            Checkpoint('2021/1/1', 100)])).summary_table()
        self.assertEqual([], perf_table.str_list())

    def test_summary_table_1month(self):
        checkpoints = [
            Checkpoint('2021/1/1', 100),
            Checkpoint('2021/1/31', 200, inflow=100, outflow=50),
            Checkpoint('2021/2/1', 210)]
        perf = Performance(Timeline(checkpoints))
        # We should only return 1 period.
        self.assertEqual(
            ['1 Month'], perf._get_periods()[1])

        perf_table = perf.summary_table()
        self.assertEqual(2, len(perf_table.list()))
        # Only check basic values (these functions are unittested elsewhere)
        self.assertEqual(['1 Month', '$100.00', '$50.00'],
                         perf_table.str_list()[0][:3])
        self.assertEqual(
            ['Overall', '$100.00', '$50.00', '+$110.00', '110.0%'],
            perf_table.str_list()[1][:5])

    def test_summary_table_1year(self):
        checkpoints = [
            Checkpoint('2021/1/1', 100),
            Checkpoint('2022/2/1', 210)]
        perf = Performance(Timeline(checkpoints))
        self.assertEqual(
            ['3 Months', '6 Months', '1 Year'], perf._get_periods()[1])
        self.assertEqual(4, len(perf.summary_table().list()))

    def test_performance_get_info(self):
        checkpoints = [
            Checkpoint('2020/1/1', 1000),
            Checkpoint('2021/1/1', 500, outflow=1000),
            Checkpoint('2022/1/1', 1000)]
        info = Performance(Timeline(checkpoints)).get_info(None, None)

        self.assertRegex(info, r'Start date +2020/01/01')
        self.assertRegex(info, r'End date +2022/01/01')
        self.assertRegex(info, r'Begin.+ \$1,000.00')
        self.assertRegex(info, r'Ending.+ \$1,000\.00')
        self.assertRegex(info, r'Inflows + \$0')
        self.assertRegex(info, r'Outflows + \$1,000\.00')
        self.assertRegex(info, r'Portfolio growth +\+\$0\.00')
        self.assertRegex(info, r'Market growth +\+\$1,000\.00')
        self.assertRegex(info, r'Portfolio growth \% +0\.0%')
        self.assertRegex(info, r'Internal.+61\.6%')
